@@ -14,30 +14,36 @@
 """Test discovery
 """
 
-import re
 import os
-import unittest
+import re
 import sys
+import unittest
 
 import six
+
+import zope.testrunner.debug
 import zope.testrunner.feature
 import zope.testrunner.layer
-import zope.testrunner.debug
 from zope.testrunner.filter import build_filtering_func
 
+
 identifier = re.compile(r'[_a-zA-Z]\w*$').match
 
 
+class DuplicateTestIDError(Exception):
+    """Raised whenever a test ID is encountered twice during loading."""
+
+
 class StartUpFailure(unittest.TestCase):
     """Empty test case added to the test suite to indicate import failures.
 
     >>> class Options(object):
     ...    post_mortem = False
     >>> options = Options()
-    
+
     Normally the StartUpFailure just acts as an empty test suite to satisfy
     the test runner and statistics:
-    
+
     >>> s = StartUpFailure(options, 'fauxmodule', None)
     >>> s
     <StartUpFailure module=fauxmodule>
@@ -85,7 +91,7 @@ class StartUpFailure(unittest.TestCase):
 
     To simulate the user pressing 'c' and hitting return in the
     debugger, we use a FakeInputContinueGenerator:
-    
+
     >>> from zope.testrunner.runner import FakeInputContinueGenerator
     >>> old_stdin = sys.stdin
     >>> sys.stdin = FakeInputContinueGenerator()
@@ -93,11 +99,12 @@ class StartUpFailure(unittest.TestCase):
     Now we can see the EndRun exception that is raised by the
     postmortem debugger to indicate that debugging is finished and the
     test run should be terminated:
-    
+
     >>> from zope.testrunner.interfaces import EndRun
     >>> try: #doctest: +ELLIPSIS
     ...   try: # try...except...finally doesn't work in Python 2.4
-    ...     print("Result:") # Needed to prevent the result from starting with '...'
+    ...     # Needed to prevent the result from starting with '...'
+    ...     print("Result:")
     ...     StartUpFailure(options, None, exc_info)
     ...   except EndRun:
     ...     print("EndRun raised")
@@ -159,6 +166,7 @@ def find_tests(options, found_suites=None):
     """
     remove_stale_bytecode(options)
     suites = {}
+    dupe_ids = set()
 
     test_accept = build_filtering_func(options.test)
     module_accept = build_filtering_func(options.module)
@@ -167,11 +175,21 @@ def find_tests(options, found_suites=None):
         found_suites = find_suites(options, accept=module_accept)
     for suite in found_suites:
         for test, layer_name in tests_from_suite(suite, options,
-                                                 accept=test_accept):
+                                                 accept=test_accept,
+                                                 duplicated_test_ids=dupe_ids):
+            if dupe_ids:
+                # If there are any duplicated test IDs, we stop trying to
+                # load tests; we'll raise an error later on with all the
+                # duplicates in it.
+                continue
             suite = suites.get(layer_name)
             if not suite:
                 suite = suites[layer_name] = unittest.TestSuite()
             suite.addTest(test)
+    if dupe_ids:
+        message_lines = ['Duplicate test IDs found:'] + sorted(dupe_ids)
+        message = '\n  '.join(message_lines)
+        raise DuplicateTestIDError(message)
     return suites
 
 
@@ -194,11 +212,12 @@ def find_suites(options, accept=None):
                     module = import_name(module_name)
                 except KeyboardInterrupt:
                     raise
-                except:
+                except BaseException:
                     exc_info = sys.exc_info()
                     if not options.post_mortem:
                         # Skip a couple of frames
-                        exc_info = exc_info[:2] + (exc_info[2].tb_next.tb_next,)
+                        exc_info = (
+                            exc_info[:2] + (exc_info[2].tb_next.tb_next,))
                     suite = StartUpFailure(
                         options, module_name, exc_info)
                 else:
@@ -206,7 +225,8 @@ def find_suites(options, accept=None):
                         if hasattr(module, options.suite_name):
                             suite = getattr(module, options.suite_name)()
                         else:
-                            suite = unittest.defaultTestLoader.loadTestsFromModule(module)
+                            loader = unittest.defaultTestLoader
+                            suite = loader.loadTestsFromModule(module)
                             if suite.countTestCases() == 0:
                                 raise TypeError(
                                     "Module %s does not define any tests"
@@ -215,13 +235,20 @@ def find_suites(options, accept=None):
                         if isinstance(suite, unittest.TestSuite):
                             check_suite(suite, module_name)
                         else:
-                            raise TypeError(
+                            # We extract the error message expression into a
+                            # local variable because we want the `raise`
+                            # statement to fit on a single line, to make the
+                            # testrunner-debugging-import-failure.rst doctest
+                            # see the same pdb output on Python 3.8 as on older
+                            # Python versions.
+                            bad_test_suite_msg = (
                                 "Invalid test_suite, %r, in %s"
                                 % (suite, module_name)
-                                )
+                            )
+                            raise TypeError(bad_test_suite_msg)
                     except KeyboardInterrupt:
                         raise
-                    except:
+                    except BaseException:
                         exc_info = sys.exc_info()
                         if not options.post_mortem:
                             # Suppress traceback
@@ -349,6 +376,8 @@ def walk_with_symlinks(options, dir):
 
 
 compiled_suffixes = '.pyc', '.pyo'
+
+
 def remove_stale_bytecode(options):
     if options.keepbytecode:
         return
@@ -369,6 +398,7 @@ def remove_stale_bytecode(options):
                                         % fullname)
                     os.unlink(fullname)
 
+
 def contains_init_py(options, fnamelist):
     """Return true iff fnamelist contains a suitable spelling of __init__.py.
 
@@ -398,7 +428,8 @@ def import_name(name):
 
 def tests_from_suite(suite, options, dlevel=1,
                      dlayer=zope.testrunner.layer.UnitTests,
-                     accept=None):
+                     accept=None,
+                     seen_test_ids=None, duplicated_test_ids=None):
     """Returns a sequence of (test, layer_name)
 
     The tree of suites is recursively visited, with the most specific
@@ -410,6 +441,14 @@ def tests_from_suite(suite, options, dlevel=1,
     accept is a function, returning boolean for given test name (see also
     build_filtering_func()).
     """
+    # We use this to track the test IDs that have been registered.
+    # tests_from_suite will complain if it encounters the same test ID
+    # twice.
+    if seen_test_ids is None:
+        seen_test_ids = set()
+    if duplicated_test_ids is None:
+        duplicated_test_ids = set()
+
     level = getattr(suite, 'level', dlevel)
     layer = getattr(suite, 'layer', dlayer)
     if not isinstance(layer, six.string_types):
@@ -418,12 +457,20 @@ def tests_from_suite(suite, options, dlevel=1,
     if isinstance(suite, unittest.TestSuite):
         for possible_suite in suite:
             for r in tests_from_suite(possible_suite, options, level, layer,
-                                      accept=accept):
+                                      accept=accept,
+                                      seen_test_ids=seen_test_ids,
+                                      duplicated_test_ids=duplicated_test_ids):
                 yield r
     elif isinstance(suite, StartUpFailure):
         yield (suite, None)
     else:
-        if level <= options.at_level:
+        if options.require_unique_ids:
+            suite_id = str(suite)
+            if suite_id in seen_test_ids:
+                duplicated_test_ids.add(suite_id)
+            else:
+                seen_test_ids.add(suite_id)
+        if options.at_level <= 0 or level <= options.at_level:
             if accept is None or accept(str(suite)):
                 yield (suite, layer)
 
@@ -450,6 +497,7 @@ def check_suite(suite, module_name):
 
 _layer_name_cache = {}
 
+
 def name_from_layer(layer):
     """Determine a name for the Layer using the namespace to avoid conflicts.
 
@@ -472,9 +520,9 @@ class Find(zope.testrunner.feature.Feature):
 
     def global_setup(self):
         # Add directories to the path
-        for path in self.runner.options.path:
+        for path in reversed(self.runner.options.path):
             if path not in sys.path:
-                sys.path.append(path)
+                sys.path.insert(0, path)
 
         tests = find_tests(self.runner.options, self.runner.found_suites)
         self.import_errors = tests.pop(None, None)
@@ -487,4 +535,3 @@ class Find(zope.testrunner.feature.Feature):
     def report(self):
         self.runner.options.output.modules_with_import_problems(
             self.import_errors)
-
