@@ -15,14 +15,21 @@
 """
 from __future__ import print_function
 
+
+try:
+    from collections.abc import MutableMapping
+except ImportError:
+    from collections import MutableMapping
+
 import doctest
 import os
-import sys
 import re
+import sys
 import tempfile
 import traceback
-
-from datetime import datetime, timedelta
+from contextlib import contextmanager
+from datetime import datetime
+from datetime import timedelta
 
 from zope.testrunner.exceptions import DocTestFailureException
 
@@ -63,8 +70,9 @@ class OutputFormatter(object):
     progress = property(lambda self: self.options.progress)
     verbose = property(lambda self: self.options.verbose)
     in_subprocess = property(
-        lambda self: self.options.resume_layer is not None and
-                     self.options.processes > 1)
+        lambda self: (
+            self.options.resume_layer is not None and
+            self.options.processes > 1))
 
     def compute_max_width(self):
         """Try to determine the terminal width."""
@@ -168,10 +176,10 @@ class OutputFormatter(object):
     def summary(self, n_tests, n_failures, n_errors, n_seconds,
                 n_skipped=0):
         """Summarize the results of a single test layer."""
-        print ("  Ran %s tests with %s failures, %s errors and "
-               "%s skipped in %s."
-               % (n_tests, n_failures, n_errors, n_skipped,
-                   self.format_seconds(n_seconds)))
+        print("  Ran %s tests with %s failures, %s errors and "
+              "%s skipped in %s."
+              % (n_tests, n_failures, n_errors, n_skipped,
+                  self.format_seconds(n_seconds)))
 
     def totals(self, n_tests, n_failures, n_errors, n_seconds,
                n_skipped=0):
@@ -206,6 +214,16 @@ class OutputFormatter(object):
             print(test)
             print("New thread(s):", new_threads)
 
+    def test_cycles(self, test, cycles):
+        """Report cyclic garbage left behind by a test."""
+        if cycles:
+            print("The following test left cyclic garbage behind:")
+            print(test)
+            for i, cy in enumerate(cycles):
+                print("Cycle", i + 1)
+                for oi in cy:
+                    print(" * ", "\n   ".join(oi))
+
     def refcounts(self, rc, prev):
         """Report a change in reference counts."""
         print("  sys refcount=%-8d change=%-6d" % (rc, rc - prev))
@@ -322,7 +340,7 @@ class OutputFormatter(object):
         sys.stdout.write(s)
         self.test_width += len(s) + 1
 
-    def test_error(self, test, seconds, exc_info):
+    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that an error occurred while running a test.
 
         Should be called right after start_test().
@@ -333,9 +351,10 @@ class OutputFormatter(object):
             print(" (%s)" % self.format_seconds_short(seconds))
         print()
         self.print_traceback("Error in test %s" % test, exc_info)
+        self.print_std_streams(stdout, stderr)
         self.test_width = self.last_width = 0
 
-    def test_failure(self, test, seconds, exc_info):
+    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that a test failed.
 
         Should be called right after start_test().
@@ -346,6 +365,7 @@ class OutputFormatter(object):
             print(" (%s)" % self.format_seconds_short(seconds))
         print()
         self.print_traceback("Failure in test %s" % test, exc_info)
+        self.print_std_streams(stdout, stderr)
         self.test_width = self.last_width = 0
 
     def print_traceback(self, msg, exc_info):
@@ -354,6 +374,21 @@ class OutputFormatter(object):
         print(msg)
         print(self.format_traceback(exc_info))
 
+    def print_std_streams(self, stdout, stderr):
+        """Emit contents of buffered standard streams."""
+        if stdout:
+            sys.stdout.write("Stdout:\n")
+            sys.stdout.write(stdout)
+            if not stdout.endswith("\n"):
+                sys.stdout.write("\n")
+            sys.stdout.write("\n")
+        if stderr:
+            sys.stderr.write("Stderr:\n")
+            sys.stderr.write(stderr)
+            if not stderr.endswith("\n"):
+                sys.stderr.write("\n")
+            sys.stderr.write("\n")
+
     def format_traceback(self, exc_info):
         """Format the traceback."""
         v = exc_info[1]
@@ -372,8 +407,12 @@ class OutputFormatter(object):
             tb = "".join(traceback.format_exception(*exc_info))
         return tb
 
-    def stop_test(self, test):
+    def stop_test(self, test, gccount):
         """Clean up the output state after a test."""
+        if gccount and self.verbose:
+            s = "!" if self.verbose == 1 else " [%d]" % gccount
+            self.test_width += len(s)
+            print(s, end="")
         if self.progress:
             self.last_width = self.test_width
         elif self.verbose > 1:
@@ -416,7 +455,12 @@ def tigetnum(attr, default=None):
             # You get curses.error when $TERM is set to an unknown name
             pass
         else:
-            return curses.tigetnum(attr)
+            try:
+                return curses.tigetnum(attr)
+            except expected_exceptions:
+                # You get TypeError on PyPy3 due to a bug:
+                # https://bitbucket.org/pypy/pypy/issue/2016/pypy3-cursestigetnum-raises-ctype
+                pass
     return default
 
 
@@ -470,12 +514,14 @@ class ColorfulOutputFormatter(OutputFormatter):
                   '?': 'character-diffs',
                   '@': 'diff-chunk',
                   '*': 'diff-chunk',
-                  '!': 'actual-output',}
+                  '!': 'actual-output',
+                  }
 
     prefixes = [('dark', '0;'),
                 ('light', '1;'),
                 ('bright', '1;'),
-                ('bold', '1;'),]
+                ('bold', '1;'),
+                ]
 
     colorcodes = {'default': 0, 'normal': 0,
                   'black': 30,
@@ -487,10 +533,10 @@ class ColorfulOutputFormatter(OutputFormatter):
                   'cyan': 36,
                   'grey': 37, 'gray': 37, 'white': 37}
 
-    slow_test_threshold = 10.0 # seconds
+    slow_test_threshold = 10.0  # seconds
 
     def color_code(self, color):
-        """Convert a color description (e.g. 'lightgray') to a terminal code."""
+        """Convert a color description (e.g. 'lightred') to a terminal code."""
         prefix_code = ''
         for prefix, code in self.prefixes:
             if color.startswith(prefix):
@@ -663,7 +709,8 @@ class ColorfulOutputFormatter(OutputFormatter):
                     print(line)
             elif line.startswith('    ') or line.strip() == '':
                 if colorize_diff and len(line) > 4:
-                    color = self.diff_color.get(line[4], color_of_indented_text)
+                    color = self.diff_color.get(
+                        line[4], color_of_indented_text)
                     print(self.colorize(color, line))
                 else:
                     if line.strip() != '':
@@ -729,38 +776,106 @@ class FakeTest(object):
         return self._id
 
 
-# Conditional imports, we don't want zope.testrunner to have a hard dependency on
-# subunit.
+# Conditional imports: we don't want zope.testrunner to have a hard
+# dependency on subunit.
 try:
     import subunit
     from subunit.iso8601 import Utc
-except ImportError:
+    subunit.StreamResultToBytes
+except (ImportError, AttributeError):
     subunit = None
 
 
-# testtools is a hard dependency of subunit itself, guarding separately for
-# richer error messages.
+# testtools is a hard dependency of subunit itself, but we guard it
+# separately for richer error messages.
 try:
-    from testtools import content
-except ImportError:
-    content = None
+    import testtools
+    from testtools.content import Content
+    from testtools.content import ContentType
+    from testtools.content import content_from_file
+    from testtools.content import text_content
+    testtools.StreamToExtendedDecorator
+except (ImportError, AttributeError):
+    testtools = None
+
+
+class _RunnableDecorator(object):
+    """Permit controlling the runnable annotation on tests.
+
+    This decorates a StreamResult, adding a setRunnable context manager to
+    indicate whether a test is runnable.  (A context manager is unidiomatic
+    here, but it's just about the simplest way to stuff the relevant state
+    through the various layers of decorators involved without accidentally
+    affecting later test results.)
+    """
+
+    def __init__(self, decorated):
+        self.decorated = decorated
+        self._runnable = True
+
+    def __getattr__(self, name):
+        return getattr(self.decorated, name)
+
+    @contextmanager
+    def setRunnable(self, runnable):
+        orig_runnable = self._runnable
+        try:
+            self._runnable = runnable
+            yield
+        finally:
+            self._runnable = orig_runnable
+
+    def status(self, **kwargs):
+        kwargs = dict(kwargs)
+        kwargs['runnable'] = self._runnable
+        self.decorated.status(**kwargs)
+
+
+class _SortedDict(MutableMapping, object):
+    """A dict that always returns items in sorted order.
+
+    This differs from collections.OrderedDict in that it returns items in
+    *sorted* order, not in insertion order.
+
+    We use this as a workaround for the fact that
+    testtools.ExtendedToStreamDecorator doesn't sort the details dict when
+    encoding it, which makes it difficult to write stable doctests for
+    subunit v2 output.
+    """
+
+    def __init__(self, items):
+        self._dict = dict(items)
+
+    def __getitem__(self, key):
+        return self._dict[key]
+
+    def __setitem__(self, key, value):
+        self._dict[key] = value
+
+    def __delitem__(self, key):
+        del self._dict[key]
+
+    def __iter__(self):
+        return iter(sorted(self._dict))
+
+    def __len__(self):
+        return len(self._dict)
 
 
 class SubunitOutputFormatter(object):
     """A subunit output formatter.
 
-    This output formatter generates subunit compatible output (see
-    https://launchpad.net/subunit). Subunit output is essentially a stream of
-    results of unit tests. In this formatter, non-test events (such as layer
-    set up) are encoded as specially tagged tests and summary-generating
-    methods (such as modules_with_import_problems) deliberately do nothing.
+    This output formatter generates subunit-compatible output (see
+    https://launchpad.net/subunit).  Subunit output is essentially a stream
+    of results of unit tests.
 
-    In particular, for a layer 'foo', the fake tests related to layer set up
-    and tear down are tagged with 'zope:layer' and are called 'foo:setUp' and
-    'foo:tearDown'. Any tests within layer 'foo' are tagged with
-    'zope:layer:foo'.
+    In this formatter, non-test events (such as layer set up) are encoded as
+    specially-tagged tests.  In particular, for a layer 'foo', the fake
+    tests related to layer setup and teardown are tagged with 'zope:layer'
+    and are called 'foo:setUp' and 'foo:tearDown'.  Any tests within layer
+    'foo' are tagged with 'zope:layer:foo'.
 
-    Note that all tags specific to this formatter begin with 'zope:'
+    Note that all tags specific to this formatter begin with 'zope:'.
     """
 
     # subunit output is designed for computers, so displaying a progress bar
@@ -779,27 +894,31 @@ class SubunitOutputFormatter(object):
 
     def __init__(self, options, stream=None):
         if subunit is None:
-            raise Exception("Requires subunit 0.0.5 or better")
-        if content is None:
-            raise Exception("Requires testtools 0.9.2 or better")
+            raise Exception('Requires subunit 0.0.11 or better')
+        if testtools is None:
+            raise Exception('Requires testtools 0.9.30 or better')
         self.options = options
 
         if stream is None:
             stream = sys.stdout
         self._stream = stream
-        self._subunit = subunit.TestProtocolClient(self._stream)
+        self._subunit = self._subunit_factory(self._stream)
 
         # Used to track the last layer that was set up or torn down. Either
         # None or (layer_name, last_touched_time).
         self._last_layer = None
         self.UTC = Utc()
         # Content types used in the output.
-        self.TRACEBACK_CONTENT_TYPE = content.ContentType(
-            'text', 'x-traceback', dict(language='python', charset='utf8'))
-        self.PROFILE_CONTENT_TYPE = content.ContentType(
+        self.TRACEBACK_CONTENT_TYPE = ContentType(
+            'text', 'x-traceback', {'language': 'python', 'charset': 'utf8'})
+        self.PROFILE_CONTENT_TYPE = ContentType(
             'application', 'x-binary-profile')
-        self.PLAIN_TEXT = content.ContentType(
-            'text', 'plain', {'charset': 'utf8'})
+        self.PLAIN_TEXT = ContentType('text', 'plain', {'charset': 'utf8'})
+
+    @classmethod
+    def _subunit_factory(cls, stream):
+        """Return a TestResult attached to the given stream."""
+        return _RunnableDecorator(subunit.TestProtocolClient(stream))
 
     def _emit_timestamp(self, now=None):
         """Emit a timestamp to the subunit stream.
@@ -811,39 +930,54 @@ class SubunitOutputFormatter(object):
         self._subunit.time(now)
         return now
 
-    def _emit_tag(self, tag):
-        self._stream.write('tags: %s\n' % (tag,))
-
-    def _stop_tag(self, tag):
-        self._stream.write('tags: -%s\n' % (tag,))
-
     def _emit_fake_test(self, message, tag, details=None):
         """Emit a successful fake test to the subunit stream.
 
         Use this to print tagged informative messages.
         """
         test = FakeTest(message)
-        self._subunit.startTest(test)
-        self._emit_tag(tag)
-        self._subunit.addSuccess(test, details=details)
+        with self._subunit.setRunnable(False):
+            self._subunit.startTest(test)
+            self._subunit.tags([tag], [])
+            self._subunit.addSuccess(test, details=details)
+            self._subunit.stopTest(test)
 
-    def _emit_error(self, error_id, tag, exc_info):
+    def _emit_error(self, error_id, tag, exc_info, runnable=False):
         """Emit an error to the subunit stream.
 
         Use this to pass on information about errors that occur outside of
         tests.
         """
         test = FakeTest(error_id)
-        self._subunit.startTest(test)
-        self._emit_tag(tag)
-        self._subunit.addError(test, exc_info)
+        with self._subunit.setRunnable(runnable):
+            self._subunit.startTest(test)
+            self._subunit.tags([tag], [])
+            self._subunit.addError(test, exc_info)
+            self._subunit.stopTest(test)
+
+    def _emit_failure(self, failure_id, tag, exc_info):
+        """Emit an failure to the subunit stream.
+
+        Use this to pass on information about failures that occur outside of
+        tests.
+        """
+        test = FakeTest(failure_id)
+        self._subunit.addFailure(test, exc_info)
+
+    def _enter_layer(self, layer_name):
+        """Tell subunit that we are entering a layer."""
+        self._subunit.tags(['zope:layer:%s' % (layer_name,)], [])
+
+    def _exit_layer(self, layer_name):
+        """Tell subunit that we are exiting a layer."""
+        self._subunit.tags([], ['zope:layer:%s' % (layer_name,)])
 
     def info(self, message):
-        """Print an informative message, but only if verbose."""
-        # info() output is not relevant to actual test results. It only says
-        # things like "Running tests" or "Tearing down left over layers",
-        # things that are communicated already by the subunit stream. Just
-        # suppress the info() output.
+        """Print an informative message."""
+        # info() output is not relevant to actual test results.  It only
+        # says things like "Running tests" or "Tearing down left over
+        # layers", things that are communicated already by the subunit
+        # stream.  Just suppress the info() output.
         pass
 
     def info_suboptimal(self, message):
@@ -867,172 +1001,31 @@ class SubunitOutputFormatter(object):
         # Or "Can't post-mortem debug when running a layer as a subprocess!"
         self._emit_fake_test(message, self.TAG_ERROR_WITH_BANNER)
 
-    def _enter_layer(self, layer_name):
-        """Signal in the subunit stream that we are entering a layer."""
-        self._emit_tag('zope:layer:%s' % (layer_name,))
-
-    def _exit_layer(self, layer_name):
-        """Signal in the subunit stream that we are leaving a layer."""
-        self._stop_tag('zope:layer:%s' % (layer_name,))
-
-    def start_set_up(self, layer_name):
-        """Report that we're setting up a layer.
-
-        We do this by emitting a tag of the form 'layer:$LAYER_NAME'.
-        """
-        now = self._emit_timestamp()
-        self._subunit.startTest(FakeTest('%s:setUp' % (layer_name,)))
-        self._emit_tag(self.TAG_LAYER)
-        self._last_layer = (layer_name, now)
-
-    def stop_set_up(self, seconds):
-        layer_name, start_time = self._last_layer
-        self._last_layer = None
-        self._emit_timestamp(start_time + timedelta(seconds=seconds))
-        self._subunit.addSuccess(FakeTest('%s:setUp' % (layer_name,)))
-        self._enter_layer(layer_name)
-
-    def start_tear_down(self, layer_name):
-        """Report that we're tearing down a layer.
-
-        We do this by removing a tag of the form 'layer:$LAYER_NAME'.
-        """
-        self._exit_layer(layer_name)
-        now = self._emit_timestamp()
-        self._subunit.startTest(FakeTest('%s:tearDown' % (layer_name,)))
-        self._emit_tag(self.TAG_LAYER)
-        self._last_layer = (layer_name, now)
-
-    def stop_tear_down(self, seconds):
-        layer_name, start_time = self._last_layer
-        self._last_layer = None
-        self._emit_timestamp(start_time + timedelta(seconds=seconds))
-        self._subunit.addSuccess(FakeTest('%s:tearDown' % (layer_name,)))
-
-    def tear_down_not_supported(self):
-        """Report that we could not tear down a layer.
-
-        Should be called right after start_tear_down().
-        """
-        layer_name, start_time = self._last_layer
-        self._last_layer = None
-        self._emit_timestamp(datetime.now(self.UTC))
-        self._subunit.addSkip(
-            FakeTest('%s:tearDown' % (layer_name,)), "tearDown not supported")
-
-    def summary(self, n_tests, n_failures, n_errors, n_seconds,
-                n_skipped=0):
-        """Print out a summary.
-
-        Since subunit is a stream protocol format, it has no need for a
-        summary. When the stream is finished other tools can generate a
-        summary if so desired.
-        """
-        pass
-
-    def start_test(self, test, tests_run, total_tests):
-        """Report that we're about to run a test.
-
-        The next output operation should be test_success(), test_error(), or
-        test_failure().
-        """
-        self._emit_timestamp()
-        # Note that this always emits newlines, so it will function as well as
-        # other start_test implementations if we are running in a subprocess.
-        self._subunit.startTest(test)
-
-    def stop_test(self, test):
-        """Clean up the output state after a test."""
-        self._subunit.stopTest(test)
-        self._stream.flush()
-
-    def stop_tests(self):
-        """Clean up the output state after a collection of tests."""
-        self._stream.flush()
-
-    def test_success(self, test, seconds):
-        self._emit_timestamp()
-        self._subunit.addSuccess(test)
-
-    def import_errors(self, import_errors):
-        """Report test-module import errors (if any)."""
-        if not import_errors:
-            return
-        for error in import_errors:
-            self._emit_error(
-                error.module, self.TAG_IMPORT_ERROR, error.exc_info)
-
-    def modules_with_import_problems(self, import_errors):
-        """Report names of modules with import problems (if any)."""
-        # This is simply a summary method, and subunit output doesn't benefit
-        # from summaries.
-        pass
-
-    def _exc_info_to_details(self, exc_info):
-        """Translate 'exc_info' into a details dictionary usable with subunit.
-        """
-        # In an ideal world, we'd use the pre-bundled 'TracebackContent' class
-        # from testtools. However, 'OutputFormatter' contains special logic to
-        # handle errors from doctests, so we have to use that and manually
-        # create an object equivalent to an instance of 'TracebackContent'.
-        formatter = OutputFormatter(None)
-        traceback = formatter.format_traceback(exc_info)
-
-        # We have no idea if the traceback is a unicode object or a bytestring
-        # with non-ASCII characters.  We had best be careful when handling it.
-        if isinstance(traceback, unicode):
-            unicode_tb = traceback
-        else:
-            # Assume the traceback was utf-8 encoded, but still be careful.
-            unicode_tb = traceback.decode('utf8', 'replace')
-
-        return {
-            'traceback': content.Content(
-                self.TRACEBACK_CONTENT_TYPE, lambda: [unicode_tb.encode('utf8')])}
-
-    def test_error(self, test, seconds, exc_info):
-        """Report that an error occurred while running a test.
-
-        Should be called right after start_test().
-
-        The next output operation should be stop_test().
-        """
-        self._emit_timestamp()
-        details = self._exc_info_to_details(exc_info)
-        self._subunit.addError(test, details=details)
-
-    def test_failure(self, test, seconds, exc_info):
-        """Report that a test failed.
-
-        Should be called right after start_test().
-
-        The next output operation should be stop_test().
-        """
-        self._emit_timestamp()
-        details = self._exc_info_to_details(exc_info)
-        self._subunit.addFailure(test, details=details)
-
     def profiler_stats(self, stats):
         """Report profiler stats."""
-        fd, filename = tempfile.mkstemp()
+        fd, filename = tempfile.mkstemp(prefix='zope.testrunner-')
         os.close(fd)
         try:
             stats.dump_stats(filename)
-            stats_dump = open(filename)
-            try:
-                profile_content = content.Content(
-                    self.PROFILE_CONTENT_TYPE, stats_dump.read)
-                details = {'profiler-stats': profile_content}
-                # Name the test 'zope:profiler_stats' just like its tag.
-                self._emit_fake_test(
-                    self.TAG_PROFILER_STATS, self.TAG_PROFILER_STATS, details)
-            finally:
-                stats_dump.close()
+            profile_content = content_from_file(
+                filename, content_type=self.PROFILE_CONTENT_TYPE)
+            details = {'profiler-stats': profile_content}
+            # Name the test 'zope:profiler_stats' just like its tag.
+            self._emit_fake_test(
+                self.TAG_PROFILER_STATS, self.TAG_PROFILER_STATS, details)
         finally:
             os.unlink(filename)
 
+    def import_errors(self, import_errors):
+        """Report test-module import errors (if any)."""
+        if import_errors:
+            for error in import_errors:
+                self._emit_error(
+                    error.module, self.TAG_IMPORT_ERROR, error.exc_info,
+                    runnable=True)
+
     def tests_with_errors(self, errors):
-        """Report tests with errors.
+        """Report names of tests with errors (if any).
 
         Simply not supported by the subunit formatter. Fancy summary output
         doesn't make sense.
@@ -1040,13 +1033,29 @@ class SubunitOutputFormatter(object):
         pass
 
     def tests_with_failures(self, failures):
-        """Report tests with failures.
+        """Report names of tests with failures (if any).
 
         Simply not supported by the subunit formatter. Fancy summary output
         doesn't make sense.
         """
         pass
 
+    def modules_with_import_problems(self, import_errors):
+        """Report names of modules with import problems (if any)."""
+        # This is simply a summary method, and subunit output doesn't
+        # benefit from summaries.
+        pass
+
+    def summary(self, n_tests, n_failures, n_errors, n_seconds,
+                n_skipped=0):
+        """Summarize the results of a single test layer.
+
+        Since subunit is a stream protocol format, it has no need for a
+        summary. When the stream is finished other tools can generate a
+        summary if so desired.
+        """
+        pass
+
     def totals(self, n_tests, n_failures, n_errors, n_seconds, n_skipped=0):
         """Summarize the results of all layers.
 
@@ -1055,20 +1064,22 @@ class SubunitOutputFormatter(object):
         """
         pass
 
+    def _emit_exists(self, test):
+        """Emit an indication that a test exists.
+
+        With the v1 protocol, we just emit a fake success line.
+        """
+        self._subunit.addSuccess(test)
+
     def list_of_tests(self, tests, layer_name):
         """Report a list of test names."""
         self._enter_layer(layer_name)
         for test in tests:
             self._subunit.startTest(test)
-            self._subunit.addSuccess(test)
+            self._emit_exists(test)
+            self._subunit.stopTest(test)
         self._exit_layer(layer_name)
 
-    def _get_text_details(self, name, text):
-        """Get a details dictionary that just has some plain text."""
-        return {
-            name: content.Content(
-                self.PLAIN_TEXT, lambda: [text.encode('utf8')])}
-
     def garbage(self, garbage):
         """Report garbage generated by tests."""
         # XXX: Really, 'garbage', 'profiler_stats' and the 'refcounts' twins
@@ -1076,54 +1087,264 @@ class SubunitOutputFormatter(object):
         # summary information for the whole suite. However, there's no event
         # on output formatters for "everything is really finished, honest". --
         # jml, 2010-02-14
-        details = self._get_text_details('garbage', unicode(garbage))
-        self._emit_fake_test(
-            self.TAG_GARBAGE, self.TAG_GARBAGE, details)
+        details = {'garbage': text_content(unicode(garbage))}
+        self._emit_fake_test(self.TAG_GARBAGE, self.TAG_GARBAGE, details)
 
     def test_garbage(self, test, garbage):
         """Report garbage generated by a test.
 
-        Encoded in the subunit stream as a test error. Clients can filter out
-        these tests based on the tag if they don't think garbage should fail
-        the test run.
+        Encoded in the subunit stream as a test error.  Clients can filter
+        out these tests based on the tag if they don't think garbage should
+        fail the test run.
         """
         # XXX: Perhaps 'test_garbage' and 'test_threads' ought to be within
         # the output for the actual test, appended as details to whatever
         # result the test gets. Not an option with the present API, as there's
         # no event for "no more output for this test". -- jml, 2010-02-14
         self._subunit.startTest(test)
-        self._emit_tag(self.TAG_GARBAGE)
+        self._subunit.tags([self.TAG_GARBAGE], [])
         self._subunit.addError(
-            test, details=self._get_text_details('garbage', unicode(garbage)))
+            test, details={'garbage': text_content(unicode(garbage))})
+        self._subunit.stopTest(test)
 
     def test_threads(self, test, new_threads):
         """Report threads left behind by a test.
 
-        Encoded in the subunit stream as a test error. Clients can filter out
-        these tests based on the tag if they don't think left-over threads
-        should fail the test run.
+        Encoded in the subunit stream as a test error.  Clients can filter
+        out these tests based on the tag if they don't think left-over
+        threads should fail the test run.
         """
         self._subunit.startTest(test)
-        self._emit_tag(self.TAG_THREADS)
+        self._subunit.tags([self.TAG_THREADS], [])
         self._subunit.addError(
-            test, details=self._get_text_details('garbage', unicode(new_threads)))
+            test, details={'threads': text_content(unicode(new_threads))})
+        self._subunit.stopTest(test)
+
+    def test_cycles(self, test, cycles):
+        """Report cycles left behind by a test."""
+        pass  # not implemented
 
     def refcounts(self, rc, prev):
         """Report a change in reference counts."""
-        details = self._get_text_details('sys-refcounts', str(rc))
-        details.update(
-            self._get_text_details('changes', str(rc - prev)))
+        details = _SortedDict({
+            'sys-refcounts': text_content(str(rc)),
+            'changes': text_content(str(rc - prev)),
+        })
         # XXX: Emit the details dict as JSON?
-        self._emit_fake_test(
-            self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
+        self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
 
     def detailed_refcounts(self, track, rc, prev):
         """Report a change in reference counts, with extra detail."""
-        details = self._get_text_details('sys-refcounts', str(rc))
-        details.update(
-            self._get_text_details('changes', str(rc - prev)))
-        details.update(
-            self._get_text_details('track', str(track.delta)))
+        details = _SortedDict({
+            'sys-refcounts': text_content(str(rc)),
+            'changes': text_content(str(rc - prev)),
+            'track': text_content(str(track.delta)),
+        })
+        self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
 
-        self._emit_fake_test(
-            self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
+    def start_set_up(self, layer_name):
+        """Report that we're setting up a layer.
+
+        We do this by emitting a fake test of the form '$LAYER_NAME:setUp'
+        and adding a tag of the form 'zope:layer:$LAYER_NAME' to the current
+        tag context.
+
+        The next output operation should be stop_set_up().
+        """
+        test = FakeTest('%s:setUp' % (layer_name,))
+        now = self._emit_timestamp()
+        with self._subunit.setRunnable(False):
+            self._subunit.startTest(test)
+            self._subunit.tags([self.TAG_LAYER], [])
+        self._last_layer = (layer_name, now)
+
+    def stop_set_up(self, seconds):
+        """Report that we've set up a layer.
+
+        Should be called right after start_set_up().
+        """
+        layer_name, start_time = self._last_layer
+        self._last_layer = None
+        test = FakeTest('%s:setUp' % (layer_name,))
+        self._emit_timestamp(start_time + timedelta(seconds=seconds))
+        with self._subunit.setRunnable(False):
+            self._subunit.addSuccess(test)
+            self._subunit.stopTest(test)
+        self._enter_layer(layer_name)
+
+    def layer_failure(self, failure_type, exc_info):
+        layer_name, start_time = self._last_layer
+        self._emit_failure(
+            '%s:%s' % (layer_name, failure_type), self.TAG_LAYER, exc_info)
+
+    def start_tear_down(self, layer_name):
+        """Report that we're tearing down a layer.
+
+        We do this by emitting a fake test of the form
+        '$LAYER_NAME:tearDown' and removing a tag of the form
+        'layer:$LAYER_NAME' from the current tag context.
+
+        The next output operation should be stop_tear_down() or
+        tear_down_not_supported().
+        """
+        test = FakeTest('%s:tearDown' % (layer_name,))
+        self._exit_layer(layer_name)
+        now = self._emit_timestamp()
+        with self._subunit.setRunnable(False):
+            self._subunit.startTest(test)
+            self._subunit.tags([self.TAG_LAYER], [])
+        self._last_layer = (layer_name, now)
+
+    def stop_tear_down(self, seconds):
+        """Report that we've torn down a layer.
+
+        Should be called right after start_tear_down().
+        """
+        layer_name, start_time = self._last_layer
+        self._last_layer = None
+        test = FakeTest('%s:tearDown' % (layer_name,))
+        self._emit_timestamp(start_time + timedelta(seconds=seconds))
+        with self._subunit.setRunnable(False):
+            self._subunit.addSuccess(test)
+            self._subunit.stopTest(test)
+
+    def tear_down_not_supported(self):
+        """Report that we could not tear down a layer.
+
+        Should be called right after start_tear_down().
+        """
+        layer_name, start_time = self._last_layer
+        self._last_layer = None
+        test = FakeTest('%s:tearDown' % (layer_name,))
+        self._emit_timestamp()
+        with self._subunit.setRunnable(False):
+            self._subunit.addSkip(test, 'tearDown not supported')
+            self._subunit.stopTest(test)
+
+    def start_test(self, test, tests_run, total_tests):
+        """Report that we're about to run a test.
+
+        The next output operation should be test_success(), test_error(), or
+        test_failure().
+        """
+        self._emit_timestamp()
+        self._subunit.startTest(test)
+
+    def test_success(self, test, seconds):
+        """Report that a test was successful.
+
+        Should be called right after start_test().
+
+        The next output operation should be stop_test().
+        """
+        self._emit_timestamp()
+        self._subunit.addSuccess(test)
+
+    def test_skipped(self, test, reason):
+        """Report that a test was skipped.
+
+        Should be called right after start_test().
+
+        The next output operation should be stop_test().
+        """
+        self._subunit.addSkip(test, reason)
+
+    def _exc_info_to_details(self, exc_info):
+        """Translate 'exc_info' into a details dict usable with subunit."""
+        # In an ideal world, we'd use the pre-bundled 'TracebackContent'
+        # class from testtools.  However, 'OutputFormatter' contains special
+        # logic to handle errors from doctests, so we have to use that and
+        # manually create an object equivalent to an instance of
+        # 'TracebackContent'.
+        formatter = OutputFormatter(None)
+        traceback = formatter.format_traceback(exc_info)
+
+        # We have no idea if the traceback is a unicode object or a
+        # bytestring with non-ASCII characters.  We had best be careful when
+        # handling it.
+        if isinstance(traceback, bytes):
+            # Assume the traceback was UTF-8-encoded, but still be careful.
+            unicode_tb = traceback.decode('utf-8', 'replace')
+        else:
+            unicode_tb = traceback
+
+        return _SortedDict({
+            'traceback': Content(
+                self.TRACEBACK_CONTENT_TYPE,
+                lambda: [unicode_tb.encode('utf8')]),
+        })
+
+    def _add_std_streams_to_details(self, details, stdout, stderr):
+        """Add buffered standard stream contents to a subunit details dict."""
+        if stdout:
+            if isinstance(stdout, bytes):
+                stdout = stdout.decode('utf-8', 'replace')
+            details['test-stdout'] = Content(
+                self.PLAIN_TEXT, lambda: [stdout.encode('utf-8')])
+        if stderr:
+            if isinstance(stderr, bytes):
+                stderr = stderr.decode('utf-8', 'replace')
+            details['test-stderr'] = Content(
+                self.PLAIN_TEXT, lambda: [stderr.encode('utf-8')])
+
+    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
+        """Report that an error occurred while running a test.
+
+        Should be called right after start_test().
+
+        The next output operation should be stop_test().
+        """
+        self._emit_timestamp()
+        details = self._exc_info_to_details(exc_info)
+        self._add_std_streams_to_details(details, stdout, stderr)
+        self._subunit.addError(test, details=details)
+
+    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
+        """Report that a test failed.
+
+        Should be called right after start_test().
+
+        The next output operation should be stop_test().
+        """
+        self._emit_timestamp()
+        details = self._exc_info_to_details(exc_info)
+        self._add_std_streams_to_details(details, stdout, stderr)
+        self._subunit.addFailure(test, details=details)
+
+    def stop_test(self, test, gccount):
+        """Clean up the output state after a test."""
+        self._subunit.stopTest(test)
+
+    def stop_tests(self):
+        """Clean up the output state after a collection of tests."""
+        # subunit handles all of this itself.
+        pass
+
+
+class SubunitV2OutputFormatter(SubunitOutputFormatter):
+    """A subunit v2 output formatter."""
+
+    @classmethod
+    def _subunit_factory(cls, stream):
+        """Return a TestResult attached to the given stream."""
+        stream_result = _RunnableDecorator(subunit.StreamResultToBytes(stream))
+        result = testtools.ExtendedToStreamDecorator(stream_result)
+        # Lift our decorating method up so that we can get at it easily.
+        result.setRunnable = stream_result.setRunnable
+        result.startTestRun()
+        return result
+
+    def error(self, message):
+        """Report an error."""
+        # XXX: Mostly used for user errors, sometimes used for errors in the
+        # test framework, sometimes used to record layer setUp failure (!!!).
+        self._subunit.status(
+            file_name='error', file_bytes=unicode(message).encode('utf-8'),
+            eof=True, mime_type=repr(self.PLAIN_TEXT))
+
+    def _emit_exists(self, test):
+        """Emit an indication that a test exists."""
+        now = datetime.now(self.UTC)
+        self._subunit.status(
+            test_id=test.id(), test_status='exists',
+            test_tags=self._subunit.current_tags, timestamp=now)
