@@ -15,32 +15,46 @@
 """
 from __future__ import print_function
 
-import optparse
-import re
+import argparse
 import os
+import re
 import sys
 
 import pkg_resources
 
-from zope.testrunner.profiling import available_profilers
-from zope.testrunner.formatter import (
-    OutputFormatter,
-    ColorfulOutputFormatter,
-    SubunitOutputFormatter,
-    )
+from zope.testrunner.formatter import ColorfulOutputFormatter
+from zope.testrunner.formatter import OutputFormatter
+from zope.testrunner.formatter import SubunitOutputFormatter
+from zope.testrunner.formatter import SubunitV2OutputFormatter
 from zope.testrunner.formatter import terminal_has_colors
+from zope.testrunner.profiling import available_profilers
+
+from .util import uses_refcounts
 
 
-parser = optparse.OptionParser("Usage: %prog [options] [MODULE] [TEST]")
+def _regex_search(s):
+    return re.compile(s).search
+
+
+parser = argparse.ArgumentParser(
+    description="Discover and run unittest tests")
+
+
+parser.add_argument("legacy_module_filter", nargs="?",
+                    help="DEPRECATED: Prefer to use --module.")
+
+
+parser.add_argument("legacy_test_filter", nargs="?",
+                    help="DEPRECATED: Prefer to use --test.")
 
 ######################################################################
 # Searching and filtering
 
-searching = optparse.OptionGroup(parser, "Searching and filtering", """\
+searching = parser.add_argument_group("Searching and filtering", """\
 Options in this group are used to define which tests to run.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--package', '--dir', '-s', action="append", dest='package',
     help="""\
 Search the given package's directories for tests.  This can be
@@ -59,7 +73,7 @@ only directories within the test search path are searched. See the
 
 """)
 
-searching.add_option(
+searching.add_argument(
     '--module', '-m', action="append", dest='module',
     help="""\
 Specify a test-module filter as a regular expression.  This is a
@@ -74,7 +88,7 @@ the test filters are searched.  If no test-module filter is specified,
 then all test modules are used.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--test', '-t', action="append", dest='test',
     help="""\
 Specify a test filter as a regular expression.  This is a
@@ -87,19 +101,19 @@ Tests matching any of the test filters are included.  If no test
 filter is specified, then all tests are run.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--unit', '-u', action="store_true", dest='unit',
     help="""\
 Run only unit tests, ignoring any layer options.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--non-unit', '-f', action="store_true", dest='non_unit',
     help="""\
 Run tests other than unit tests.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--layer', action="append", dest='layer',
     help="""\
 Specify a test layer to run.  The option can be given multiple times
@@ -114,102 +128,119 @@ layer named 'zope.testrunner.layer.UnitTests' is reserved for
 unit tests, however, take note of the --unit and non-unit options.
 """)
 
-searching.add_option(
-    '-a', '--at-level', type='int', dest='at_level',
+searching.add_argument(
+    '-a', '--at-level', type=int, dest='at_level',
+    default=1,
     help="""\
 Run the tests at the given level.  Any test at a level at or below
-this is run, any test at a level above this is not run.  Level 0
+this is run, any test at a level above this is not run.  Level <= 0
 runs all tests.
 """)
 
-searching.add_option(
+searching.add_argument(
     '--all', action="store_true", dest='all',
     help="Run tests at all levels.")
 
-searching.add_option(
+searching.add_argument(
     '--list-tests', action="store_true", dest='list_tests',
+    default=False,
     help="List all tests that matched your filters.  Do not run any tests.")
 
-parser.add_option_group(searching)
+searching.add_argument(
+    '--require-unique', action="store_true", dest='require_unique_ids',
+    default=False,
+    help="""\
+Require that all test IDs be unique and raise an error if duplicates are
+encountered.
+""")
+
 
 ######################################################################
 # Reporting
 
-reporting = optparse.OptionGroup(parser, "Reporting", """\
+reporting = parser.add_argument_group("Reporting", """\
 Reporting options control basic aspects of test-runner output
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--verbose', '-v', action="count", dest='verbose',
+    default=0,
     help="""\
 Make output more verbose.
 Increment the verbosity level.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--quiet', '-q', action="store_true", dest='quiet',
     help="""\
 Make the output minimal, overriding any verbosity options.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--progress', '-p', action="store_true", dest='progress',
     help="""\
 Output progress status
 """)
 
-reporting.add_option(
-    '--no-progress',action="store_false", dest='progress',
+reporting.add_argument(
+    '--no-progress', action="store_false", dest='progress',
     help="""\
 Do not output progress status.  This is the default, but can be used to
 counter a previous use of --progress or -p.
 """)
 
-# We use a noop callback because the actual processing will be done in the
-# get_options function, but we want optparse to generate appropriate help info
-# for us, so we add an option anyway.
-reporting.add_option(
-    '--auto-progress', action="callback", callback=lambda *args: None,
+# The actual processing will be done in the get_options function, but
+# we want argparse to generate appropriate help info for us, so we add
+# an option anyway.
+reporting.add_argument(
+    '--auto-progress', action="store_const", const=None,
     help="""\
 Output progress status, but only when stdout is a terminal.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--color', '-c', action="store_true", dest='color',
     help="""\
 Colorize the output.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--no-color', '-C', action="store_false", dest='color',
     help="""\
 Do not colorize the output.  This is the default, but can be used to
 counter a previous use of --color or -c.
 """)
 
-# We use a noop callback because the actual processing will be done in the
-# get_options function, but we want optparse to generate appropriate help info
-# for us, so we add an option anyway.
-reporting.add_option(
-    '--auto-color', action="callback", callback=lambda *args: None,
+# The actual processing will be done in the get_options function, but
+# we want argparse to generate appropriate help info for us, so we add
+# an option anyway.
+reporting.add_argument(
+    '--auto-color', action="store_const", const=None,
     help="""\
 Colorize the output, but only when stdout is a terminal.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--subunit', action="store_true", dest='subunit',
     help="""\
-Use subunit output. Will not be colorized.
+Use subunit v1 output.  Will not be colorized.
 """)
 
-reporting.add_option(
-    '--slow-test', type='float', dest='slow_test_threshold', metavar='N',
+reporting.add_argument(
+    '--subunit-v2', action="store_true", dest='subunit_v2',
+    help="""\
+Use subunit v2 output.  Will not be colorized.
+""")
+
+reporting.add_argument(
+    '--slow-test', type=float, dest='slow_test_threshold', metavar='N',
+    default=10,
     help="""\
 With -c and -vvv, highlight tests that take longer than N seconds (default:
-%default).
+%(default)s).
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '-1', '--hide-secondary-failures',
     action="store_true", dest='report_only_first_failure',
     help="""\
@@ -217,7 +248,7 @@ Report only the first failure in a doctest. (Examples after the
 failure are still executed, in case they do any cleanup.)
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--show-secondary-failures',
     action="store_false", dest='report_only_first_failure',
     help="""\
@@ -225,48 +256,70 @@ Report all failures in a doctest.  This is the default, but can
 be used to counter a default use of -1 or --hide-secondary-failures.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--ndiff', action="store_true", dest="ndiff",
     help="""\
 When there is a doctest failure, show it as a diff using the ndiff.py utility.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--udiff', action="store_true", dest="udiff",
     help="""\
 When there is a doctest failure, show it as a unified diff.
 """)
 
-reporting.add_option(
+reporting.add_argument(
     '--cdiff', action="store_true", dest="cdiff",
     help="""\
 When there is a doctest failure, show it as a context diff.
 """)
 
-parser.add_option_group(reporting)
+reporting.add_argument(
+    '--ignore-new-thread',
+    metavar='REGEXP',
+    action="append",
+    default=[],
+    dest='ignore_new_threads',
+    help="""\
+If a thread with this name is left behind, don't report this at the end.
+This is a case-sensitive regular expression, used in match mode.
+This option can be used multiple times. If a thread name matches any of them,
+it will be ignored.
+""")
+
+reporting.add_argument(
+    '--buffer', action="store_true", dest="buffer",
+    help="""\
+Buffer the standard output and standard error streams during each test.
+Output during a passing test is discarded.
+Output during failing or erroring tests is echoed.
+This option is enabled by default if --subunit or --subunit-v2 is in use, to
+avoid corrupting the subunit stream.
+""")
+
 
 ######################################################################
 # Analysis
 
-analysis = optparse.OptionGroup(parser, "Analysis", """\
+analysis = parser.add_argument_group("Analysis", """\
 Analysis options provide tools for analysing test output.
 """)
 
 
-analysis.add_option(
+analysis.add_argument(
     '--stop-on-error', '--stop', '-x', action="store_true",
     dest='stop_on_error',
     help="Stop running tests after first test failure or error."
     )
 
-analysis.add_option(
+analysis.add_argument(
     '--post-mortem', '--pdb', '-D', action="store_true", dest='post_mortem',
     help="Enable post-mortem debugging of test failures"
     )
 
 
-analysis.add_option(
-    '--gc', '-g', action="append", dest='gc', type="int",
+analysis.add_argument(
+    '--gc', '-g', action="append", dest='gc', type=int,
     help="""\
 Set the garbage collector generation threshold.  This can be used
 to stress memory and gc correctness.  Some crashes are only
@@ -278,25 +331,36 @@ Python gc_threshold settings.
 
 """)
 
-analysis.add_option(
-    '--gc-option', '-G', action="append", dest='gc_option', type="choice",
-    choices=['DEBUG_STATS', 'DEBUG_COLLECTABLE', 'DEBUG_UNCOLLECTABLE',
+analysis.add_argument(
+    '--gc-option', '-G', action="append", dest='gc_option',
+    choices={'DEBUG_STATS', 'DEBUG_COLLECTABLE', 'DEBUG_UNCOLLECTABLE',
              'DEBUG_INSTANCES', 'DEBUG_OBJECTS', 'DEBUG_SAVEALL',
-             'DEBUG_LEAK'],
+             'DEBUG_LEAK'},
     help="""\
 Set a Python gc-module debug flag.  This option can be used more than
 once to set multiple flags.
 """)
 
-analysis.add_option(
-    '--repeat', '-N', action="store", type="int", dest='repeat',
+if uses_refcounts:
+    analysis.add_argument(
+        '--gc-after-test', action="store_true", dest='gc_after_test',
+        help="""\
+    After each test, call 'gc.collect' and record the return
+    value *rv*; when *rv* is non-zero, output '!' on verbosity level 1
+    and '[*rv*]' on higher verbosity levels.\n
+    On verbosity level 4 or higher output detailed cycle information.
+    """)
+
+analysis.add_argument(
+    '--repeat', '-N', action="store", type=int, dest='repeat',
+    default=1,
     help="""\
 Repeat the tests the given number of times.  This option is used to
 make sure that tests leave their environment in the state they found
 it and, with the --report-refcounts option to look for memory leaks.
 """)
 
-analysis.add_option(
+analysis.add_argument(
     '--report-refcounts', '-r', action="store_true", dest='report_refcounts',
     help="""\
 After each run of the tests, output a report summarizing changes in
@@ -304,22 +368,22 @@ refcounts by object type.  This option that requires that Python was
 built with the --with-pydebug option to configure.
 """)
 
-analysis.add_option(
-    '--coverage', action="store", type='string', dest='coverage',
+analysis.add_argument(
+    '--coverage', action="store", dest='coverage',
     help="""\
 Perform code-coverage analysis, saving trace data to the directory
 with the given name.  A code coverage summary is printed to standard
 out.
 """)
 
-analysis.add_option(
-    '--profile', action="store", dest='profile', type="choice",
-    choices=list(available_profilers),
+analysis.add_argument(
+    '--profile', action="store", dest='profile',
+    choices=set(available_profilers),
     help="""\
-Run the tests under cProfiler or hotshot and display the top 50 stats, sorted
+Run the tests under cProfiler and display the top 50 stats, sorted
 by cumulative time and number of calls.
 """)
-analysis.add_option(
+analysis.add_argument(
     '--profile-directory', action="store", dest='prof_dir', default='.',
     help="""\
 Directory for temporary profiler files.  All files named tests_profile.*.prof
@@ -328,29 +392,17 @@ of the test runner in parallel, be sure to tell them to use different
 directories, so they won't step on each other's toes.
 """)
 
-def do_pychecker(*args):
-    if not os.environ.get("PYCHECKER"):
-        os.environ["PYCHECKER"] = "-q"
-    import pychecker.checker
-
-analysis.add_option(
-    '--pychecker', action="callback", callback=do_pychecker,
-    help="""\
-Run the tests under pychecker
-""")
-
-parser.add_option_group(analysis)
-
 ######################################################################
 # Setup
 
-setup = optparse.OptionGroup(parser, "Setup", """\
+setup = parser.add_argument_group("Setup", """\
 Setup options are normally supplied by the testrunner script, although
 they can be overridden by users.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--path', action="append", dest='path',
+    type=os.path.abspath,
     help="""\
 Specify a path to be added to Python's search path.  This option can
 be used multiple times to specify multiple search paths.  The path is
@@ -362,8 +414,9 @@ This option also specifies directories to be searched for tests.
 See the search_directory.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--test-path', action="append", dest='test_path',
+    type=os.path.abspath,
     help="""\
 Specify a path to be searched for tests, but not added to the Python
 search path.  This option can be used multiple times to specify
@@ -373,16 +426,16 @@ although it can be overridden by users.  Only tests found in the path
 will be run.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--package-path', action="append", dest='package_path', nargs=2,
+    metavar="ARG",
     help="""\
 Specify a path to be searched for tests, but not added to the Python
 search path.  Also specify a package for files found in this path.
 This is used to deal with directories that are stitched into packages
 that are not otherwise searched for tests.
 
-This option takes 2 arguments.  The first is a path name. The second is
-the package name.
+This option takes 2 arguments specifying the path and the package.
 
 This option can be used multiple times to specify
 multiple search paths.  The path is usually specified by the
@@ -391,8 +444,10 @@ although it can be overridden by users.  Only tests found in the path
 will be run.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--tests-pattern', action="store", dest='tests_pattern',
+    default=_regex_search('^tests$'),
+    type=_regex_search,
     help="""\
 The test runner looks for modules containing tests.  It uses this
 pattern to identify these modules.  The modules may be either packages
@@ -403,58 +458,62 @@ test-file-pattern to identify python files within the package
 containing tests.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--suite-name', action="store", dest='suite_name',
+    default='test_suite',
     help="""\
 Specify the name of the object in each test_module that contains the
 module's test suite.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--test-file-pattern', action="store", dest='test_file_pattern',
+    default=_regex_search('^test'),
+    type=_regex_search,
     help="""\
 Specify a pattern for identifying python files within a tests package.
 See the documentation for the --tests-pattern option.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--ignore_dir', action="append", dest='ignore_dir',
+    default=['.git', '.svn', 'CVS', '{arch}', '.arch-ids', '_darcs'],
     help="""\
 Specifies the name of a directory to ignore when looking for tests.
 """)
 
-setup.add_option(
+setup.add_argument(
     '--shuffle', action="store_true", dest='shuffle',
     help="""\
 Shuffles the order in which tests are ran.
 """)
 
-setup.add_option(
-    '--shuffle-seed', action="store", dest='shuffle_seed', type="int",
+setup.add_argument(
+    '--shuffle-seed', action="store", dest='shuffle_seed', type=int,
     help="""\
 Value used to initialize the tests shuffler. Specify a value to create
 repeatable random ordered tests.
 """)
 
-parser.add_option_group(setup)
 
 ######################################################################
 # Other
 
-other = optparse.OptionGroup(parser, "Other", "Other options")
+other = parser.add_argument_group("Other", "Other options")
 
-other.add_option(
+other.add_argument(
     '--version', action="store_true", dest='showversion',
     help="Print the version of the testrunner, and exit.")
 
-other.add_option(
-    '-j', action="store", type="int", dest='processes',
+other.add_argument(
+    '-j', action="store", type=int, dest='processes',
+    default=1,
     help="""\
 Use up to given number of parallel processes to execute tests.  May decrease
-test run time substantially.  Defaults to %default.
+test run time substantially.  Defaults to %(default)s.
 """)
 
-other.add_option(
+other.add_argument(
     '--keepbytecode', '-k', action="store_true", dest='keepbytecode',
     help="""\
 Normally, the test runner scans the test paths and the test
@@ -466,7 +525,7 @@ scan.  If you know you haven't removed any modules since last running
 the tests, can make the test run go much faster.
 """)
 
-other.add_option(
+other.add_argument(
     '--usecompiled', action="store_true", dest='usecompiled',
     help="""\
 Normally, a package must contain an __init__.py file, and only .py files
@@ -479,31 +538,12 @@ tests against a tree where the .py files have been removed after
 compilation to .pyc/.pyo.  Use of this option implies --keepbytecode.
 """)
 
-other.add_option(
+other.add_argument(
     '--exit-with-status', action="store_true", dest='exitwithstatus',
     help="""DEPRECATED: The test runner will always exit with a status.\
 """)
 
 
-parser.add_option_group(other)
-
-######################################################################
-# Default values
-
-parser.set_defaults(
-    ignore_dir=['.git', '.svn', 'CVS', '{arch}', '.arch-ids', '_darcs'],
-    tests_pattern='^tests$',
-    at_level=1,
-    test_file_pattern='^test',
-    suite_name='test_suite',
-    list_tests=False,
-    slow_test_threshold=10,
-    processes=1,
-    verbose=0,
-    repeat=1,
-    )
-
-
 ######################################################################
 # Command-line processing
 
@@ -514,6 +554,7 @@ def merge_options(options, defaults):
         if (value is not None) and (odict[name] is None):
             odict[name] = value
 
+
 def get_options(args=None, defaults=None):
     # Because we want to inspect stdout and decide to colorize or not, we
     # replace the --auto-color option with the appropriate --color or
@@ -546,15 +587,14 @@ def get_options(args=None, defaults=None):
     apply_auto_progress(defaults)
 
     if defaults:
-        defaults, _ = parser.parse_args(defaults)
-        assert not _
+        defaults = parser.parse_args(defaults)
     else:
         defaults = None
 
     if args is None:
         args = sys.argv
 
-    options, positional = parser.parse_args(args[1:], defaults)
+    options = parser.parse_args(args[1:], defaults)
     options.original_testrunner_args = args
 
     if options.showversion:
@@ -563,9 +603,10 @@ def get_options(args=None, defaults=None):
         options.fail = True
         return options
 
-    if options.subunit:
+    if options.subunit or options.subunit_v2:
         try:
             import subunit
+            subunit
         except ImportError:
             print("""\
         Subunit is not installed. Please install Subunit
@@ -573,8 +614,19 @@ def get_options(args=None, defaults=None):
         """)
             options.fail = True
             return options
-        else:
-            options.output = SubunitOutputFormatter(options)
+        options.buffer = True
+
+    if options.subunit and options.subunit_v2:
+        print("""\
+        You may only use one of --subunit and --subunit-v2.
+        """)
+        options.fail = True
+        return options
+
+    if options.subunit:
+        options.output = SubunitOutputFormatter(options)
+    elif options.subunit_v2:
+        options.output = SubunitV2OutputFormatter(options)
     elif options.color:
         options.output = ColorfulOutputFormatter(options)
         options.output.slow_test_threshold = options.slow_test_threshold
@@ -583,32 +635,28 @@ def get_options(args=None, defaults=None):
 
     options.fail = False
 
-    if positional:
-        module_filter = positional.pop(0)
+    if options.legacy_module_filter:
+        module_filter = options.legacy_module_filter
         if module_filter != '.':
             if options.module:
                 options.module.append(module_filter)
             else:
                 options.module = [module_filter]
 
-        if positional:
-            test_filter = positional.pop(0)
+        if options.legacy_test_filter:
+            test_filter = options.legacy_test_filter
             if options.test:
                 options.test.append(test_filter)
             else:
                 options.test = [test_filter]
 
-            if positional:
-                parser.error("Too many positional arguments")
+    options.ignore_dir = set(options.ignore_dir)
+    options.test = options.test or ['.']
+    module_set = bool(options.module)
+    options.module = options.module or ['.']
 
-    options.ignore_dir = dict([(d,1) for d in options.ignore_dir])
-    options.test_file_pattern = re.compile(options.test_file_pattern).search
-    options.tests_pattern = re.compile(options.tests_pattern).search
-    options.test = [t for t in options.test or ('.')]
-    options.module = [m for m in options.module or ('.')]
-
-    options.path = [os.path.abspath(p) for p in options.path or ()]
-    options.test_path = [os.path.abspath(p) for p in options.test_path or ()]
+    options.path = options.path or []
+    options.test_path = options.test_path or []
     options.test_path += options.path
 
     options.test_path = ([(path, '') for path in options.test_path]
@@ -644,7 +692,7 @@ def get_options(args=None, defaults=None):
         # XXX Argh.
         options.layer = ['zope.testrunner.layer.UnitTests']
 
-    options.layer = options.layer and dict([(l, 1) for l in options.layer])
+    options.layer = options.layer and {layer: 1 for layer in options.layer}
 
     if options.usecompiled:
         options.keepbytecode = options.usecompiled
@@ -661,7 +709,6 @@ def get_options(args=None, defaults=None):
         options.fail = True
         return options
 
-
     if options.report_refcounts and not hasattr(sys, "gettotalrefcount"):
         print("""\
         The Python you are running was not configured
@@ -671,9 +718,19 @@ def get_options(args=None, defaults=None):
         options.fail = True
         return options
 
+    if module_set and options.require_unique_ids:
+        # We warn if --module and --require-unique are specified at the same
+        # time, though we don't exit.
+        print("""\
+        You specified a module along with --require-unique;
+        --require-unique will not try to enforce test ID uniqueness when
+        working with a specific module.
+        """)
+
     return options
 
-def normalize_package(package, package_map={}):
+
+def normalize_package(package, package_map=None):
     r"""Normalize package name passed to the --package option.
 
         >>> normalize_package('zope.testrunner')
@@ -704,6 +761,7 @@ def normalize_package(package, package_map={}):
         'zope.something.tests'
 
     """
+    package_map = {} if package_map is None else package_map
     package = package.replace('\\', '/')
     if package.endswith('/'):
         package = package[:-1]
