@@ -15,64 +15,65 @@
 """
 from __future__ import print_function
 
-import collections
-import subprocess
 import errno
 import gc
-import inspect
+import io
+import os
+import pprint
 import re
+import subprocess
 import sys
 import threading
 import time
 import traceback
-try:
-    import unittest2 as unittest
-except ImportError:
-    import unittest
+import unittest
+import warnings
+from contextlib import contextmanager
 
 from six import StringIO
-from zope.testrunner.find import import_name
-from zope.testrunner.find import name_from_layer, _layer_name_cache
-from zope.testrunner.layer import UnitTests
-from zope.testrunner.layer import EmptyLayer
-from zope.testrunner.layer import EmptySuite
-from zope.testrunner.options import get_options
-from zope.testrunner.refcount import TrackRefs
+
 import zope.testrunner
-import zope.testrunner.coverage
 import zope.testrunner._doctest
-import zope.testrunner.logsupport
-import zope.testrunner.selftest
-import zope.testrunner.profiling
+import zope.testrunner.coverage
+import zope.testrunner.debug
 import zope.testrunner.filter
 import zope.testrunner.garbagecollection
-import zope.testrunner.listing
-import zope.testrunner.statistics
-import zope.testrunner.process
 import zope.testrunner.interfaces
-import zope.testrunner.debug
-import zope.testrunner.tb_format
+import zope.testrunner.listing
+import zope.testrunner.logsupport
+import zope.testrunner.process
+import zope.testrunner.profiling
+import zope.testrunner.selftest
 import zope.testrunner.shuffle
+import zope.testrunner.statistics
+import zope.testrunner.tb_format
+from zope.testrunner import threadsupport
+from zope.testrunner.find import _layer_name_cache
+from zope.testrunner.find import import_name
+from zope.testrunner.find import name_from_layer
+from zope.testrunner.layer import EmptyLayer
+from zope.testrunner.layer import EmptySuite
+from zope.testrunner.layer import UnitTests
+from zope.testrunner.options import get_options
+from zope.testrunner.refcount import TrackRefs
+
+from .digraph import DiGraph
+from .util import is_jython
+from .util import uses_refcounts
+
 
 try:
-    import Queue # Python 2
+    import Queue  # Python 2
 except ImportError:
     # Python 3
-    import queue as Queue # Python 3
-
-try:
-    from unittest import _UnexpectedSuccess # Python 3.1
-except ImportError:
-    try:
-        from unittest.case import _UnexpectedSuccess # Python 2.7 and 3.2
-    except ImportError:
-        class _UnexpectedSuccess(Exception):
-            pass
+    import queue as Queue  # Python 3
 
 
-PYREFCOUNT_PATTERN = re.compile('\[[0-9]+ refs\]')
+class UnexpectedSuccess(Exception):
+    pass
 
-is_jython = sys.platform.startswith('java')
+
+PYREFCOUNT_PATTERN = re.compile(r'\[[0-9]+ refs\]')
 
 
 class SubprocessError(Exception):
@@ -98,10 +99,20 @@ class Runner(object):
     executing tests as well as configuring itself from the (command-line)
     options passed into it.
 
+    .. versionchanged:: 4.8.0
+       Add the *warnings* keyword argument. If this is ``None`` (the default)
+       and the user hasn't configured Python otherwise with command-line
+       arguments or environment variables, we will enable the default
+       warnings, including ``DeprecationWarning``, when running tests.
+       Otherwise, it can be any string acceptable to
+       :func:`warnings.simplefilter` and that filter will be in effect while
+       running tests.
+
     """
 
     def __init__(self, defaults=None, args=None, found_suites=None,
-                 options=None, script_parts=None, cwd=None):
+                 options=None, script_parts=None, cwd=None,
+                 warnings=None):
         if defaults is None:
             self.defaults = []
         else:
@@ -112,6 +123,18 @@ class Runner(object):
         self.script_parts = script_parts
         self.cwd = cwd
         self.failed = True
+        if warnings is None and not sys.warnoptions:
+            # even if DeprecationWarnings are ignored by default
+            # print them anyway unless other warnings settings are
+            # specified by the warnings arg or the -W python flag
+            self.warnings = 'default'
+        else:
+            # here self.warnings is set either to the value passed
+            # to the warnings args or to None.
+            # If the user didn't pass a value self.warnings will
+            # be None. This means that the behavior is unchanged
+            # and depends on the values passed to -W.
+            self.warnings = warnings
 
         self.ran = 0
         self.skipped = []
@@ -154,32 +177,37 @@ class Runner(object):
         self.layer_name_cache = _layer_name_cache
         self.layer_name_cache.clear()
 
-        # Global setup
-        for feature in self.features:
-            feature.global_setup()
+        with self._enabled_warnings():
+            # Enable warnings during setup so that
+            # warnings raised on import (which we do for test
+            # discover) can be reported.
 
-        # Late setup
-        #
-        # Some system tools like profilers are really bad with stack frames.
-        # E.g. hotshot doesn't like it when we leave the stack frame that we
-        # called start() from.
-        for feature in self.features:
-            feature.late_setup()
-
-        try:
-            if self.do_run_tests:
-                self.run_tests()
-        finally:
-            # Early teardown
-            for feature in reversed(self.features):
-                feature.early_teardown()
-            # Global teardown
-            for feature in reversed(self.features):
-                feature.global_teardown()
-
-        if self.show_report:
+            # Global setup
             for feature in self.features:
-                feature.report()
+                feature.global_setup()
+
+            # Late setup
+            #
+            # Some system tools like profilers are really bad with stack
+            # frames.  E.g. hotshot doesn't like it when we leave the stack
+            # frame that we called start() from.
+            for feature in self.features:
+                feature.late_setup()
+
+            try:
+                if self.do_run_tests:
+                    self.run_tests()
+            finally:
+                # Early teardown
+                for feature in reversed(self.features):
+                    feature.early_teardown()
+                # Global teardown
+                for feature in reversed(self.features):
+                    feature.global_teardown()
+
+            if self.show_report:
+                for feature in self.features:
+                    feature.report()
 
     def configure(self):
         if self.args is None:
@@ -233,6 +261,27 @@ class Runner(object):
         # Remove all features that aren't activated
         self.features = [f for f in self.features if f.active]
 
+    @contextmanager
+    def _enabled_warnings(self):
+        """
+        A context manager to enable warnings as configured.
+        """
+        with warnings.catch_warnings():
+            if self.warnings:
+                # if self.warnings is set, use it to filter all the warnings
+                warnings.simplefilter(self.warnings)
+                # if the filter is 'default' or 'always', special-case the
+                # warnings from the deprecated unittest methods to show them
+                # no more than once per module, because they can be fairly
+                # noisy.  The -Wd and -Wa flags can be used to bypass this
+                # only when self.warnings is None.
+                if self.warnings in ['default', 'always']:
+                    warnings.filterwarnings(
+                        'module',
+                        category=DeprecationWarning,
+                        message=r'Please use assert\w+ instead.')
+            yield
+
     def run_tests(self):
         """Run all tests that were registered.
 
@@ -264,6 +313,9 @@ class Runner(object):
                 should_resume = True
                 break
 
+            if self.options.stop_on_error and (self.failures or self.errors):
+                break
+
         if should_resume:
             if layers_to_run:
                 self.ran += resume_tests(
@@ -274,11 +326,22 @@ class Runner(object):
         if setup_layers:
             if self.options.resume_layer is None:
                 self.options.output.info("Tearing down left over layers:")
-            tear_down_unneeded(self.options, (), setup_layers, True)
+            tear_down_unneeded(
+                self.options, (), setup_layers, self.errors, optional=True)
 
         self.failed = bool(self.import_errors or self.failures or self.errors)
 
 
+def handle_layer_failure(failure_type, output, errors):
+    if hasattr(output, 'layer_failure'):
+        output.layer_failure(failure_type.subunit_label, sys.exc_info())
+    else:
+        f = StringIO()
+        traceback.print_exc(file=f)
+        output.error(f.getvalue())
+    errors.append((failure_type, sys.exc_info()))
+
+
 def run_tests(options, tests, name, failures, errors, skipped, import_errors):
     repeat = options.repeat or 1
     repeat_range = iter(range(repeat))
@@ -321,7 +384,9 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
                         test.debug()
                     except KeyboardInterrupt:
                         raise
-                    except:
+                    except unittest.SkipTest as e:
+                        result.addSkip(test, str(e))
+                    except BaseException:
                         result.addError(
                             test,
                             sys.exc_info()[:2] + (sys.exc_info()[2].tb_next, ),
@@ -346,17 +411,16 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
         t = time.time() - t
         output.stop_tests()
         failures.extend(result.failures)
+        n_failures = len(result.failures)
         if hasattr(result, 'unexpectedSuccesses'):
             # Python versions prior to 2.7 do not have the concept of
             # unexpectedSuccesses.
             failures.extend(result.unexpectedSuccesses)
-        if not hasattr(result, 'skipped'):
-            # Only in Python >= 2.7, Python >= 3.1, and when using unittest2
-            result.skipped = []
+            n_failures += len(result.unexpectedSuccesses)
         skipped.extend(result.skipped)
         errors.extend(result.errors)
         output.summary(n_tests=result.testsRun,
-                       n_failures=len(failures),
+                       n_failures=n_failures,
                        n_errors=len(result.errors) + len(import_errors),
                        n_seconds=t,
                        n_skipped=len(result.skipped))
@@ -399,10 +463,10 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
     output = options.output
     gathered = []
     gather_layers(layer, gathered)
-    needed = dict([(l, 1) for l in gathered])
+    needed = dict([(ly, 1) for ly in gathered])
     if options.resume_number != 0:
         output.info("Running %s tests:" % layer_name)
-    tear_down_unneeded(options, needed, setup_layers)
+    tear_down_unneeded(options, needed, setup_layers, errors)
 
     if options.resume_layer is not None:
         output.info_suboptimal("  Running in a subprocess.")
@@ -411,11 +475,10 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
         setup_layer(options, layer, setup_layers)
     except zope.testrunner.interfaces.EndRun:
         raise
+    except MemoryError:
+        raise
     except Exception:
-        f = StringIO()
-        traceback.print_exc(file=f)
-        output.error(f.getvalue())
-        errors.append((SetUpLayerFailure(layer), sys.exc_info()))
+        handle_layer_failure(SetUpLayerFailure(layer), output, errors)
         return 0
     else:
         return run_tests(options, tests, layer_name, failures, errors, skipped,
@@ -424,6 +487,8 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
 
 class SetUpLayerFailure(unittest.TestCase):
 
+    subunit_label = 'setUp'
+
     def __init__(self, layer):
         super(SetUpLayerFailure, self).__init__()
         self.layer = layer
@@ -432,15 +497,29 @@ class SetUpLayerFailure(unittest.TestCase):
         pass
 
     def __str__(self):
-        return "Layer: %s" % (name_from_layer(self.layer))
+        return "Layer: %s.setUp" % (name_from_layer(self.layer))
 
 
+class TearDownLayerFailure(unittest.TestCase):
+
+    subunit_label = 'tearDown'
+
+    def __init__(self, layer):
+        super(TearDownLayerFailure, self).__init__()
+        self.layer = layer
+
+    def runTest(self):
+        pass
+
+    def __str__(self):
+        return "Layer: %s.tearDown" % (name_from_layer(self.layer))
+
 
 def spawn_layer_in_subprocess(result, script_parts, options, features,
                               layer_name, layer, failures, errors, skipped,
                               resume_number, cwd=None):
     output = options.output
-
+    child = None
     try:
         # BBB
         if script_parts is None:
@@ -455,7 +534,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
         # this is because of a bug in Python (http://www.python.org/sf/900092)
         if (options.profile == 'hotshot'
-            and sys.version_info[:3] <= (2, 4, 1)):
+                and sys.version_info[:3] <= (2, 4, 1)):
             args.insert(1, '-O')
 
         debugargs = args  # save them before messing up for windows
@@ -467,10 +546,22 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
         for feature in features:
             feature.layer_setup(layer)
 
-        child = subprocess.Popen(args, shell=False, stdin=subprocess.PIPE,
+        child = subprocess.Popen(
+            args, shell=False, stdin=subprocess.PIPE,
             stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd,
             close_fds=not sys.platform.startswith('win'))
 
+        def reader_thread(f, buf):
+            buf.append(f.read())
+
+        # Start reading stderr in a thread.  This means we don't hang if the
+        # subprocess writes more to stderr than the pipe capacity.
+        stderr_buf = []
+        stderr_thread = threading.Thread(
+            target=reader_thread, args=(child.stderr, stderr_buf))
+        stderr_thread.daemon = True
+        stderr_thread.start()
+
         while True:
             try:
                 while True:
@@ -480,10 +571,10 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                     # return its lines as a batch). We don't want too much
                     # buffering because this foils automatic and human monitors
                     # trying to verify that the subprocess is still alive.
-                    l = child.stdout.readline()
-                    if not l:
+                    line = child.stdout.readline()
+                    if not line:
                         break
-                    result.write(l)
+                    result.write(line)
             except IOError as e:
                 if e.errno == errno.EINTR:
                     # If the subprocess dies before we finish reading its
@@ -496,8 +587,9 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             else:
                 break
 
-        # Now stderr should be ready to read the whole thing.
-        errlines = child.stderr.read().splitlines()
+        # Now we should be able to finish reading stderr.
+        stderr_thread.join()
+        errlines = stderr_buf[0].splitlines()
         erriter = iter(errlines)
         nfail = nerr = 0
         for line in erriter:
@@ -513,7 +605,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             if options.verbose >= 1:
                 errmsg += "\nChild command line: %s" % debugargs
             if (options.verbose >= 2 or
-                (options.verbose == 1 and len(errlines) < 20)):
+                    (options.verbose == 1 and len(errlines) < 20)):
                 errmsg += ("\nChild stderr was:\n" +
                            "\n".join("  " + line.decode('utf-8', 'replace')
                                      for line in errlines))
@@ -526,7 +618,6 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                                      for line in errlines[-10:]))
             output.error_with_banner(errmsg)
 
-
         while nfail > 0:
             nfail -= 1
             # Doing erriter.next().strip() confuses the 2to3 fixer, so
@@ -544,6 +635,33 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
     finally:
         result.done = True
+        if child is not None:
+            # Regardless of whether the process ran to completion, we
+            # must properly cleanup the process to avoid
+            # `ResourceWarning: subprocess XXX is still alive` and
+            # `ResourceWarning: unclosed file` for its stdout and
+            # stderr.
+            child.kill()
+            child.communicate()
+
+
+def _get_output_buffer(stream):
+    """Get a binary-safe version of a stream."""
+    try:
+        fileno = stream.fileno()
+    except (io.UnsupportedOperation, AttributeError):
+        pass
+    else:
+        # Win32 mangles \r\n to \n and that breaks streams.  See
+        # https://bugs.launchpad.net/bugs/505078.
+        if sys.platform == 'win32':
+            import msvcrt
+            msvcrt.setmode(fileno, os.O_BINARY)
+    try:
+        stream.write(b'')
+    except TypeError:
+        return stream.buffer
+    return stream
 
 
 class AbstractSubprocessResult(object):
@@ -565,8 +683,6 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
     """Keeps stdout around for later processing,"""
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if not _is_dots(out):
             self.stdout.append(out)
 
@@ -574,18 +690,19 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
 class ImmediateSubprocessResult(AbstractSubprocessResult):
     """Sends complete output to queue."""
 
+    def __init__(self, layer_name, queue):
+        super(ImmediateSubprocessResult, self).__init__(layer_name, queue)
+        self.stream = _get_output_buffer(sys.stdout)
+
     def write(self, out):
-        if not isinstance(out, str):
-            # In Python 3, a Popen process stdout uses bytes,
-            # While normal stdout uses strings. So we need to convert
-            # from bytes to strings here.
-            out = out.decode()
-        sys.stdout.write(out)
+        self.stream.write(out)
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
+        self.stream.flush()
+
+
+_is_dots = re.compile(br'\.+(\r\n?|\n)').match  # Windows sneaks in a \r\n.
 
 
-_is_dots = re.compile(r'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
 class KeepaliveSubprocessResult(AbstractSubprocessResult):
     "Keeps stdout for later processing; sends marks to queue to show activity."
 
@@ -598,8 +715,6 @@ class KeepaliveSubprocessResult(AbstractSubprocessResult):
     done = property(lambda self: self._done, _set_done)
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if _is_dots(out):
             self.queue.put((self.layer_name, out.strip()))
         else:
@@ -612,7 +727,8 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     stdout_queue = None
     if options.processes == 1:
         result_factory = ImmediateSubprocessResult
-    elif options.verbose > 1:
+    elif (options.verbose > 1 and
+            not options.subunit and not options.subunit_v2):
         result_factory = KeepaliveSubprocessResult
         stdout_queue = Queue.Queue()
     else:
@@ -634,6 +750,8 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     current_result = next(results_iter)
     last_layer_intermediate_output = None
     output = None
+    # Get an object that (only) accepts bytes
+    stdout = _get_output_buffer(sys.stdout)
     while ready_threads or running_threads:
         while len(running_threads) < options.processes and ready_threads:
             thread = ready_threads.pop(0)
@@ -641,7 +759,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             running_threads.append(thread)
 
         for index, thread in reversed(list(enumerate(running_threads))):
-            if not thread.isAlive():
+            if not thread.is_alive():
                 del running_threads[index]
 
         # Clear out any messages in queue
@@ -654,18 +772,21 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             if layer_name != last_layer_intermediate_output:
                 # Clarify what layer is reporting activity.
                 if previous_output is not None:
-                    sys.stdout.write(']\n')
-                sys.stdout.write(
-                    '[Parallel tests running in %s:\n  ' % (layer_name,))
+                    stdout.write(b']\n')
+                stdout.write(
+                    ('[Parallel tests running in '
+                     '%s:\n  ' % (layer_name,)).encode('utf-8'))
                 last_layer_intermediate_output = layer_name
-            sys.stdout.write(output)
+            if not isinstance(output, bytes):
+                output = output.encode('utf-8')
+            stdout.write(output)
         # Display results in the order they would have been displayed, had the
         # work not been done in parallel.
         while current_result and current_result.done:
             if output is not None:
-                sys.stdout.write(']\n')
+                stdout.write(b']\n')
                 output = None
-            sys.stdout.writelines(current_result.stdout)
+            stdout.writelines(current_result.stdout)
 
             try:
                 current_result = next(results_iter)
@@ -673,35 +794,40 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
                 current_result = None
 
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
-        time.sleep(0.01) # Keep the loop from being too tight.
+        stdout.flush()
+        time.sleep(0.01)  # Keep the loop from being too tight.
 
     # Return the total number of tests run.
     return sum(r.num_ran for r in results)
 
 
-def tear_down_unneeded(options, needed, setup_layers, optional=False):
+def tear_down_unneeded(options, needed, setup_layers, errors, optional=False):
     # Tear down any layers not needed for these tests. The unneeded layers
     # might interfere.
-    unneeded = [l for l in setup_layers if l not in needed]
+    unneeded = [layer for layer in setup_layers if layer not in needed]
     unneeded = order_by_bases(unneeded)
     unneeded.reverse()
     output = options.output
-    for l in unneeded:
-        output.start_tear_down(name_from_layer(l))
+    for layer in unneeded:
+        output.start_tear_down(name_from_layer(layer))
         t = time.time()
         try:
             try:
-                if hasattr(l, 'tearDown'):
-                    l.tearDown()
+                if hasattr(layer, 'tearDown'):
+                    layer.tearDown()
             except NotImplementedError:
                 output.tear_down_not_supported()
                 if not optional:
-                    raise CanNotTearDown(l)
+                    raise CanNotTearDown(layer)
+            except MemoryError:
+                raise
+            except Exception:
+                handle_layer_failure(
+                    TearDownLayerFailure(layer), output, errors)
             else:
                 output.stop_tear_down(time.time() - t)
         finally:
-            del setup_layers[l]
+            del setup_layers[layer]
 
 
 cant_pm_in_subprocess_message = """
@@ -722,6 +848,8 @@ def setup_layer(options, layer, setup_layers):
         if hasattr(layer, 'setUp'):
             try:
                 layer.setUp()
+            except MemoryError:
+                raise
             except Exception:
                 if options.post_mortem:
                     if options.resume_layer:
@@ -753,6 +881,10 @@ class TestResult(unittest.TestResult):
         for test in tests:
             count += test.countTestCases()
         self.count = count
+        self._stdout_buffer = None
+        self._stderr_buffer = None
+        self._original_stdout = sys.stdout
+        self._original_stderr = sys.stderr
 
     def testSetUp(self):
         """A layer may define a setup method to be called before each
@@ -774,29 +906,78 @@ class TestResult(unittest.TestResult):
             if hasattr(layer, 'testTearDown'):
                 layer.testTearDown()
 
+    def _makeBufferedStdStream(self):
+        """Make a buffered stream to replace a standard stream."""
+        if sys.version_info[0] >= 3:
+            # The returned stream needs to have a 'buffer' attribute, since
+            # some tests may expect that to exist on standard streams, and a
+            # 'getvalue' method for the convenience of _restoreStdStreams.
+            # This requires some care.
+            class BufferedStandardStream(io.TextIOWrapper):
+                def getvalue(self):
+                    return self.buffer.getvalue().decode(
+                        encoding=self.encoding, errors=self.errors)
+
+            return BufferedStandardStream(
+                io.BytesIO(), newline='\n', write_through=True)
+        else:
+            return StringIO()
+
+    def _setUpStdStreams(self):
+        """Set up buffered standard streams, if requested."""
+        if self.options.buffer:
+            if self._stdout_buffer is None:
+                self._stdout_buffer = self._makeBufferedStdStream()
+            if self._stderr_buffer is None:
+                self._stderr_buffer = self._makeBufferedStdStream()
+            sys.stdout = self._stdout_buffer
+            sys.stderr = self._stderr_buffer
+
+    def _restoreStdStreams(self):
+        """Restore the buffered standard streams and return any contents."""
+        if self.options.buffer:
+            stdout = sys.stdout.getvalue()
+            stderr = sys.stderr.getvalue()
+            sys.stdout = self._original_stdout
+            sys.stderr = self._original_stderr
+            self._stdout_buffer.seek(0)
+            self._stdout_buffer.truncate(0)
+            self._stderr_buffer.seek(0)
+            self._stderr_buffer.truncate(0)
+            return stdout, stderr
+        else:
+            return None, None
+
     def startTest(self, test):
+        self._test_state = test.__dict__.copy()
         self.testSetUp()
         unittest.TestResult.startTest(self, test)
-        testsRun = self.testsRun - 1 # subtract the one the base class added
+        testsRun = self.testsRun - 1  # subtract the one the base class added
         count = test.countTestCases()
         self.testsRun = testsRun + count
 
         self.options.output.start_test(test, self.testsRun, self.count)
 
-        self._threads = threading.enumerate()
+        self._threads = threadsupport.enumerate()
         self._start_time = time.time()
 
+        self._setUpStdStreams()
+
     def addSuccess(self, test):
+        self._restoreStdStreams()
         t = max(time.time() - self._start_time, 0.0)
         self.options.output.test_success(test, t)
 
     def addSkip(self, test, reason):
+        self._restoreStdStreams()
         unittest.TestResult.addSkip(self, test, reason)
         self.options.output.test_skipped(test, reason)
 
     def addError(self, test, exc_info):
+        stdout, stderr = self._restoreStdStreams()
         self.options.output.test_error(test, time.time() - self._start_time,
-                                       exc_info)
+                                       exc_info,
+                                       stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addError(self, test, exc_info)
 
@@ -811,8 +992,10 @@ class TestResult(unittest.TestResult):
             self.stop()
 
     def addFailure(self, test, exc_info):
+        stdout, stderr = self._restoreStdStreams()
         self.options.output.test_failure(test, time.time() - self._start_time,
-                                         exc_info)
+                                         exc_info,
+                                         stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addFailure(self, test, exc_info)
 
@@ -824,14 +1007,18 @@ class TestResult(unittest.TestResult):
             self.stop()
 
     def addExpectedFailure(self, test, exc_info):
+        self._restoreStdStreams()
         t = max(time.time() - self._start_time, 0.0)
         self.options.output.test_success(test, t)
 
         unittest.TestResult.addExpectedFailure(self, test, exc_info)
 
     def addUnexpectedSuccess(self, test):
-        self.options.output.test_error(test, time.time() - self._start_time,
-                                       (_UnexpectedSuccess, None, None))
+        stdout, stderr = self._restoreStdStreams()
+        self.options.output.test_error(
+            test, time.time() - self._start_time,
+            (UnexpectedSuccess, UnexpectedSuccess(), None),
+            stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addUnexpectedSuccess(self, test)
 
@@ -842,13 +1029,40 @@ class TestResult(unittest.TestResult):
                                                       " as a subprocess!")
             else:
                 # XXX: what exc_info? there's no exc_info!
-                zope.testrunner.debug.post_mortem(exc_info)
+                # flake8 is correct, but keep it quiet for now ...
+                zope.testrunner.debug.post_mortem(exc_info)  # noqa: F821
         elif self.options.stop_on_error:
             self.stop()
 
     def stopTest(self, test):
         self.testTearDown()
-        self.options.output.stop_test(test)
+        # Without clearing, cyclic garbage referenced by the test
+        # would be reported in the following test.
+        test.__dict__.clear()
+        test.__dict__.update(self._test_state)
+        del self._test_state
+        cycles = None
+        if (uses_refcounts and self.options.gc_after_test and
+                self.options.verbose >= 4):
+            gc_opts = gc.get_debug()
+            gc.set_debug(gc.DEBUG_SAVEALL)
+            gc.collect()
+            if gc.garbage:
+                g = DiGraph(gc.garbage)
+                for obj in gc.garbage:
+                    g.add_neighbors(obj, gc.get_referents(obj))
+                cycles = [[repr_lines(o) for o in c] for c in g.sccs()]
+                del gc.garbage[:]
+                g = obj = None  # avoid to hold cyclic garbage
+                # Python 2: avoid to hold cyclic garbage
+                o = c = None  # noqa: F841
+            gc.set_debug(gc_opts)
+        gccount = gc.collect() \
+            if uses_refcounts and self.options.gc_after_test else 0
+        self.options.output.stop_test(test, gccount)
+
+        if cycles:
+            self.options.output.test_cycles(test, cycles)
 
         if is_jython:
             pass
@@ -859,10 +1073,13 @@ class TestResult(unittest.TestResult):
                 #       printed for every subsequent test.
 
         # Did the test leave any new threads behind?
-        new_threads = [t for t in threading.enumerate()
-                         if (t.isAlive()
-                             and
-                             t not in self._threads)]
+        new_threads = []
+        for t in threadsupport.enumerate():
+            if t.is_alive() and t not in self._threads:
+                if not any([re.match(p, t.name)
+                            for p in self.options.ignore_new_threads]):
+                    new_threads.append(t)
+
         if new_threads:
             self.options.output.test_threads(test, new_threads)
 
@@ -902,20 +1119,26 @@ def layer_sort_key(layer):
 
     # Note: we cannot reuse gather_layers() here because it uses a
     # different traversal order.
+    binding = {}  # hack to avoid recursion -- for PY2
+
     def _gather(layer):
         seen.add(layer)
         # We make the simplifying assumption that the order of initialization
         # of base layers does not matter.  Given that, traversing the bases
         # in reverse order here keeps the ordering of layers in
-        # testrunner-layers.txt the same as it was in older versions of
+        # testrunner-layers.rst the same as it was in older versions of
         # zope.testrunner, so let's use that.
         for base in layer.__bases__[::-1]:
             if base is not object and base not in seen:
-                _gather(base)
+                binding["_gather"](base)
         key.append(layer)
 
+    binding["_gather"] = _gather
     _gather(layer)
-    return tuple(name_from_layer(l) for l in key if l != UnitTests)
+    try:
+        return tuple(name_from_layer(ly) for ly in key if ly != UnitTests)
+    finally:
+        binding.clear()  # break reference cycle
 
 
 def order_by_bases(layers):
@@ -951,8 +1174,31 @@ class FakeInputContinueGenerator:
     def readline(self):
         print('c\n')
         print('*'*70)
-        print ("Can't use pdb.set_trace when running a layer"
-               " as a subprocess!")
+        print("Can't use pdb.set_trace when running a layer"
+              " as a subprocess!")
         print('*'*70)
         print()
         return 'c\n'
+
+    def close(self):
+        pass
+
+
+def repr_lines(obj, max_width=75, max_lines=5):
+    """represent *obj* by a sequence of text lines.
+
+    Use at most *max_lines*, each with at most *max_width* chars.
+    """
+    try:
+        oi = pprint.pformat(obj)
+    except Exception:
+        # unprintable
+        oi = "%s instance at 0x%x" % (obj.__class__, id(obj))
+    # limit
+    cmps = oi.split("\n", max_lines)
+    if len(cmps) > max_lines:
+        cmps[-1] = "..."
+    for i, li in enumerate(cmps):
+        if len(li) > max_width:
+            cmps[i] = li[:max_width - 3] + "..."
+    return cmps
